#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import urlparse
import Queue
import threading
import sys
import datetime
import calendar

import logging

from netsa.data.times import bin_datetime
import netsa.json as json


class FrankError(Exception):
    # I dunno, something funky happened.
    pass

class AlertListener(object):
    # The callback used to forward alerts to the AlertWriter
    def __init__(self, queue):
        self._queue = queue
        self._log = logging.getLogger("pipevis.listenerapp")
    def __call__(self, alert) :
        self._log.debug("Putting alert %s in queue %s", alert, self._queue)
        self._queue.put(alert)



class AlertSource(object):
    def __init__(self, alert_factory, callback, origin, channel_name):
        # sink_factory == Snarf Sink class (or another thing
        # implementing that interface)
        self._sink_factory = alert_factory
        self._callback = callback
        self._origin = origin
        self._channel_name = channel_name
        self._sink = None
    def start(self):
        log = logging.getLogger('alertsource')
        log.debug(self._sink_factory)
        log.debug("callback: %s", str(self._callback))
        log.debug("origin: %s", self._origin)
        log.debug("channel: %s", self._channel_name)
        self._sink = self._sink_factory(callback=self._callback,
                                        origin=self._origin,
                                        channel=self._channel_name)
    def stop(self):
        log = logging.getLogger('alertsource')
        log.debug("stopping sink")
        self._sink.stop()
        log.debug("sink stopped")
        self._sink = None
        


def to_epoch(dt):
    return calendar.timegm(dt.timetuple())


def parse_alert_tags(tags):
    # Extract analysis type and name from alert tag
    def extract(prefix):
        try:
            return [t for t in tags if t.startswith(prefix)][0][len(prefix)+1:]
        except IndexError:
            return "UNSUPPLIED"

    rc = extract("type"), extract("name")
    return rc



class WriterOperations(object):
    qstr_insert_alert = """
        INSERT into alerts (
               generator,
               generator_version,
               timestamp,
               severity,
               type,
               name,
               extra_info)
        VALUES (
               :generator,
               :generator_version,
               :timestamp,
               :severity,
               :type,
               :name,
               :extra_info)
    """
    qstr_insert_bin = """
        INSERT INTO alert_vol (timestamp, num_alerts, type, name)
            SELECT :timestamp AS timestamp, 0 AS num_alerts,
                   :type as type, :name as name
            WHERE NOT EXISTS (SELECT * FROM alert_vol
                      WHERE timestamp=:timestamp
                      AND name=:name
                      AND type=:type)
    """
    qstr_update_bin = """
        UPDATE alert_vol SET num_alerts = num_alerts + 1
        WHERE timestamp = :timestamp
        AND type=:type
        AND name=:name
    """
    #def __init__(self, db_conn, q_insert_alert, q_insert_bin, q_update_bin):
    def __init__(self, db_query, db_conn):
        self._db_conn = db_conn
        self._q_insert_alert = db_query(self.qstr_insert_alert)
        self._q_insert_bin = db_query(self.qstr_insert_bin)
        self._q_update_bin = db_query(self.qstr_update_bin)
    def insert_alert(self, alert, alert_type, alert_name):
        self._q_insert_alert(self._db_conn,
                             generator=alert.get_generator(),
                             generator_version=alert.get_generator_version(),
                             timestamp=to_epoch(alert.get_timestamp()),
                             severity=alert.get_severity().severity,
                             type=alert_type,
                             name=alert_name,
                             extra_info=json.dumps(alert.get_fields_json()))
    def insert_bin(self, bin_timestamp, alert_type, alert_name):
        self._q_insert_bin(self._db_conn,
                           timestamp=to_epoch(bin_timestamp),
                           type=alert_type,
                           name=alert_name)
    def update_bin(self, bin_timestamp, alert_type, alert_name):
        self._q_update_bin(self._db_conn,
                           timestamp=to_epoch(bin_timestamp),
                           type=alert_type,
                           name=alert_name)
    def commit(self): self._db_conn.commit()
        
        



class SqlWriter(object):
    # Sends Pipeline alerts to a SQL database
    def __init__(self, writeroperations, bin_size):
        self._writeroperations = writeroperations
        if isinstance(bin_size, datetime.timedelta):
            self._bin_size = bin_size
        else:
            self._bin_size = datetime.timedelta(seconds=bin_size)
        # db: the netsa.sql module, or something that implements its
        # interface (again, for testing)
    def __call__(self, alert):
        bin_timestamp = bin_datetime(dt=self._bin_size,
                                     t=alert.get_timestamp())
        alert_type, alert_name = parse_alert_tags(alert.get_tags())
        self._writeroperations.insert_alert(alert=alert,
                                            alert_type=alert_type,
                                            alert_name=alert_name)
        self._writeroperations.insert_bin(bin_timestamp=bin_timestamp,
                                          alert_type=alert_type,
                                          alert_name=alert_name)
        self._writeroperations.update_bin(bin_timestamp=bin_timestamp,
                                          alert_type=alert_type,
                                          alert_name=alert_name)
        self._writeroperations.commit()
        



# Listener main loop

class MainLoop(object):
    """
    Loop until told to stop, doing work by calling a processor
    function. The loop can be told to stop either internally in the
    *processor* function or externally by calling the :meth:`stop` method

    *initializer* is a function that takes no arguments and prepares
    for processing. *finalizer* is run when processing is
    complete. *processor* is run repeatedly to execute
    processing.

    The return values of *initializer* and *finalizer* are
    ignored. However, exceptions raised from them (especially
    *initializer* will halt processing and propagate
    normally.

    *processor* may return one of the following values::

       :attribute:`MainLoop.READY`

          Processing has been completed for the moment, and the
          processor is ready for more work.
          
       :attribute:`MainLoop.DONE`
       
          Processing is complete, and the loop should move on to
          calling *finalizer* and shutting down.

    If *processor* returns any value other than those just documented,
    it is treated as an error. Exceptions raised in *processor* also
    propogate normally.
    """

    READY = object()
    DONE  = object()

    FINISHED = object()
    STOPPED = object()
    
    def __init__(self,
                 initializer,
                 processor,
                 finalizer,
                 poll_sleep):
        self._initializer = initializer
        self._processor = processor
        self._finalizer = finalizer
        self._poll_sleep = poll_sleep
        self._cond_wakeup = threading.Condition()
        self._external_stop = False

    
    def __call__(self):
        """
        Start the processing loop. This function will block until
        processing is complete. The function returns one of the
        following values::

           :attribute:`MainLoop.FINISHED`

              The loop was stopped from within, possibly because no
              more work remained to be done. (Or because work could no
              longer be performed.)

           :attribute:`MainLoop.STOPPED`

              The loop was stopped externally by calling the
              :meth:`MainLoop.stop` method.

        Any unhandled exceptions during processing will be raised.
        """
        self._initializer()
        self._cond_wakeup.acquire()
        try:
            while True:
                if self._external_stop:
                    break
                else:
                    proc_rv = self._processor()
                    if proc_rv == self.READY:
                        self._cond_wakeup.wait(timeout=self._poll_sleep)
                    elif proc_rv == self.DONE:
                        break
                    else:
                        raise ValueError(
                            "processor returned "
                            "unexpected value '%s'" % str(proc_rv))
            logging.debug("finalizing")
            self._finalizer()
            if self._external_stop:
                return self.STOPPED
            else:
                return self.FINISHED
        finally:
            logging.debug("all done")
            self._cond_wakeup.release()
            

    def stop(self):
        self._cond_wakeup.acquire()
        try:
            self._external_stop = True
            self._cond_wakeup.notify()
        finally:
            self._cond_wakeup.release()



# Object creation and wiring things together

try:
    import snarf
    capability_snarf = True
except ImportError:
    # We should be able to run the tests without the alert library
    # installed. Therefore, keep going. However, make a note that the
    # library is absent, in case someone cares.
    capability_snarf = False

try:
    import netsa.sql
    capability_netsa_sql = True
except ImportError:
    # Similarly, we should be able to run (at least some) tests
    # without the netsa.sql library. Note it and move on
    capability_netsa_sql = False
    

class Mother(object):
    def new_queue(self):
        return Queue.Queue()
    def new_alert_listener(self, queue):
        return AlertListener(queue=queue)
    def new_alert_source(self, origin, channel, callback):
        return AlertSource(alert_factory=snarf.Sink,
                           callback=callback,
                           origin=origin, 
                           channel_name=channel)
    def new_sql_writer(self, connection_uri, bin_size):
        db_query = netsa.sql.db_query
        db_conn = netsa.sql.db_connect(connection_uri)
        ops = WriterOperations(db_query=db_query, db_conn=db_conn)
        return SqlWriter(writeroperations=ops, bin_size=bin_size)
    def new_listener_functions(self, source, listener, queue, writer):
        return get_listener_functions(
            source=source,
            listener=listener,
            queue=queue,
            writer=writer)
    def new_main_loop(self, initializer, processor, finalizer, poll_sleep):
        return MainLoop(
            initializer=initializer,
            processor=processor,
            finalizer=finalizer,
            poll_sleep=poll_sleep)






class ListenerApp(object):
    # Does the work of the listener application, starting and stopping
    # the event loop. Note that the run() method blocks; if that's
    # undesirable, it should go in a thread.
    def __init__(self, opts, mother=None):
        if mother is None:
            mother = Mother()
        queue = mother.new_queue()
        listener = mother.new_alert_listener(queue=queue)
        source = mother.new_alert_source(callback=listener,
                                         origin=opts['origin'], 
                                         channel=opts['channel'])
                                         
        writer = mother.new_sql_writer(connection_uri=opts['dst'],
                                       bin_size=300)
        log = logging.getLogger("pipevis.listenerapp")
        def initialize():
            source.start()
            log.debug("Listener initialized")
        def process():
            log.debug("processing alerts")
            try:
                while True:
                    alert = queue.get(block=False)
                    # Note: queue.get() raises Empty if there's
                    # nothing in queue, so this is never reached.
                    writer(alert)
                    log.debug("alert written")
            except Queue.Empty:
                log.debug("processing complete")
                return MainLoop.READY
        def finalize():
            source.stop()
            log.debug("listener finalized")

        poll_sleep = opts['poll_sleep']
        poll_sleep = int(poll_sleep) / 1000

        self._loop = mother.new_main_loop(
            initializer=initialize,
            processor=process,
            finalizer=finalize,
            poll_sleep=poll_sleep)
    
    def run(self):
        self._loop()

    def stop(self):
        self._loop.stop()
    




