
# Tools for Python applications running as Unix services

import sys
import os
import errno
import pwd
import signal
import time
import socket
import atexit

from netsa.files import acquire_pidfile_lock, release_pidfile_lock

# setproctitle is a third-party lib, so only support it if it's
# available
try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = lambda t: None

# Many thanks to http://www.itp.uzh.ch/~dpotter/howto/daemonize

class WontRunAsRoot(Exception):
    # User is root or wants to daemonize a process as root, but
    # run_as_root is False
    pass

class ChildForkFailure(Exception):
    # Child process failed to initialize correctly.
    pass

def daemonize(service_name=None,
              lockfile=None,
              lockfile_dir="/var/run",
              cwd="/",
              run_as=None,
              run_as_root=False):
    """
    Spawns a child process with no controlling terminal which will run
    any code following the call to :func:`daemonize`.

    :func:`daemonize` makes the following changes to the environment:
    
       * ``stdin``, ``stdout`` and ``stderr`` is redirected to
         ``dev/null``

       * Terminal-related signals (SIGTSTP, SIGTTIN, SIGTTOU, SIGHUP)
         are ignored.

       * The working directory is changed to the value of *cwd*

       * The running user is set to *run_as* 
    
    *service_name* is used to determine the name of artifacts
    associated with this service, such as the lock file. By default,
    the service name will be the value of ``sys.argv[0]``.

    This function will attempt to create a lock file named *lockfile*
    (default: ``<service_name>.pid``) in *lockfile_dir* (default:
    ``var/run/``). This file will contain the process ID of the
    daemonized process. If a lock file already exists at this location
    and contains an active PID (or otherwise cannot be created), this
    function will raise an exception.

    *cwd* specifies the working directory of the daemon process. By
    default, the working directory will be set to the root (``/``).

    By default, the daemonized process will run as the current
    user. If the running user has the appropriate permissions, the
    daemon process will run as the user name specified in *run_as* (by
    default, the current user). This function will refuse to run a
    process as the root user unless *run_as_root* is ``True``.

    This function raises an exception in the parent process if the
    daemonizing process cannot be completed. Otherwise, it returns the
    full path of the lock file. The calling code should use this in
    conjunction with :func:`netsa.files.release_pidfile_lock` to
    ensure the lockfile is properly removed when the daemon exits.
    """
    if service_name is None:
        service_name = sys.argv[0]

    if lockfile is None:
        lockfile = "%s.pid" % service_name

    lockfile_fullpath = os.path.join(lockfile_dir, lockfile)

    # Get our PID for later, in the children
    top_pid = os.getpid()

    if run_as is None:
        # I'm not sure geteuid() is really needed here. I just don't
        # want to risk something being run with root privs when the
        # user says not to.
        if os.getuid() == 0 or os.geteuid() == 0 and not run_as_root:
            raise WontRunAsRoot()
        else:
            run_as_uid = os.getuid()
    else:
        pw = pwd.getpwnam(run_as)
        run_as_uid = pw[2]
        if run_as_uid == 0 and not run_as_root:
            raise WontRunAsRoot()

    # Set up success and failure signals between parent and child
    child_ok = [None]
    def set_child_ok(signum, frame):
        child_ok[0] = True
    signal.signal(signal.SIGUSR1, set_child_ok)

    # Try to create a socketpair that the child will use to communicate
    # error information back to the parent.
    try:
        psock, csock = socket.socketpair()
        psock.settimeout(0.0)
        csock.settimeout(0.0)
    except:
        psock, csock = (None, None)
    def read_err():
        if psock is None:
            return ""
        else:
            return psock.recv(1024)
    def set_child_fail(signum, frame):
        child_ok[0] = False
    signal.signal(signal.SIGUSR2, set_child_fail)
    

    # Fork #1
    fork1_pid = os.fork()
    if fork1_pid > 0:
        # First parent

        # Close child socket
        csock.close()

        # give child time (~2sec) to set up, then die with
        # os._exit(). (IOW, the child "inherits" any enclosing
        # exception/signal/atexit handlers that may be in effect when
        # daemonize is called.)
        for x in xrange(200):
            time.sleep(.01)
            if child_ok[0] is not None:
                break
        if not child_ok[0]:
            raise ChildForkFailure(read_err())
        else:
            os._exit(0)
    else:
        # First child - Setup comms with the parent, detach, then call
        # fork #2

        # Close parent socket
        psock.close()

        def fail_msg(msg):
            csock.sendall(msg)
            os.kill(top_pid, signal.SIGUSR2)
            os._exit(1)

        # Note that setuid, setsid and chdir, contra the Python
        # documentation (which says they have the semantics of
        # setuid(2), setsid(2) and chdir(2)), always return None and
        # raise OSError on failure.
        try:
            os.setuid(run_as_uid)
            os.umask(0)
            os.setsid()
            os.chdir(cwd)
            fork2_pid = os.fork()
        except:
            t, v, tb = sys.exc_info()
            fail_msg("%s: %s" % (t.__name__, v))

        if fork2_pid > 0:
            # Second parent.
            csock.close()
            os._exit(0)
        else:
            # Second child
            
            try:
                # Ignore terminal-based signals
                for sig in (signal.SIGTSTP,
                            signal.SIGTTOU,
                            signal.SIGTTIN,
                            signal.SIGHUP):
                    signal.signal(sig, signal.SIG_IGN)

                setproctitle(service_name)

                # Redirect output
                sys.stdout.flush()
                sys.stderr.flush()
                std = ((sys.stdin,  open('/dev/null', 'r')),
                       (sys.stdout, open('/dev/null', 'a+')),
                       (sys.stderr, open('/dev/null', 'a+', 0)))
                # (dup2 has the same semantics as setuid et al., BTW)
                for old, new in std:
                    os.dup2(new.fileno(), old.fileno())

                lock_acquired = acquire_pidfile_lock(lockfile_fullpath)
            except:
                t, v, tb = sys.exc_info()
                fail_msg("%s: %s" % (t.__name__, v))

            if not lock_acquired:
                fail_msg("Couldn't acquire pidfile lock on %s" %
                         lockfile_fullpath)

            # We're daemonized. Send the all clear to our grandparent,
            # if it's still around.
            try:
                os.kill(top_pid, signal.SIGUSR1)
            except:
                pass

            return lockfile_fullpath

    
    
            

__all__ = ["daemonize"]
