# Test the Writeroperations object, which translates high-level writer
# actions into database queries
import unittest
import calendar

from netsa.data.times import make_datetime

from fake import FakeAlert, Severity
from pipevis import WriterOperations


def db_query_factory(on_db_query_call):
    class FakeDBQuery(object):
        def __init__(self, qstr):
            self._qstr = qstr
            self._on_db_query_call = on_db_query_call
        def __call__(self, conn, **kargs):
            self._on_db_query_call(self._qstr, conn, **kargs)
    return FakeDBQuery

reference_alert = {
    "generator" : "org.cert.netsa.pipeline",
    "generator_version" : "0.1.2",
    "timestamp" : make_datetime("2011/01/01:02:14:00"),
    "severity": 1,
    "tags" : ["type:watchlist", "name:test-watch"],
    "data" : "{extra_info: 'A JSON blob of stuff'}"
}

def to_epoch(dt):
    return calendar.timegm(dt.timetuple())


class TestWriterOperations(unittest.TestCase):
    def setUp(self):
        class fake_db_conn(object):
            def __init__(self): self._committed = False
            def commit(self): self._committed = True
            def test_committed(self): return self._committed
                
        self.vals = {}
        self.db_conn = fake_db_conn()
        def on_db_query_call(qstr, conn, **kargs):
            self.assertEquals(conn, self.db_conn)
            self.assertFalse(qstr in self.vals)
            self.vals[qstr] = kargs

        self.db_query = db_query_factory(on_db_query_call)

        self.ops = WriterOperations(db_query=self.db_query,
                                    db_conn=self.db_conn)

    def test_insert_alert(self):
        alert = FakeAlert(**reference_alert)
        alert_type = "alert-type"
        alert_name = "alert-name"

        self.ops.insert_alert(alert, alert_type, alert_name)

        rc_params  = self.vals[self.ops.qstr_insert_alert]
        self.assertEquals(rc_params, dict(
                generator=reference_alert['generator'],
                generator_version=reference_alert['generator_version'],
                timestamp=to_epoch(reference_alert['timestamp']),
                severity=reference_alert['severity'],
                type=alert_type,
                name=alert_name,
                extra_info=reference_alert['data']))

    def test_insert_bin(self):
        bin_timestamp = make_datetime("2011/01/01:05:00:00")
        alert_type = "alert-type"
        alert_name = "alert-name"

        self.ops.insert_bin(bin_timestamp, alert_type, alert_name)

        rc_params = self.vals[self.ops.qstr_insert_bin]

        self.assertEquals(rc_params, dict(
                timestamp=to_epoch(bin_timestamp),
                type=alert_type,
                name=alert_name))

    def test_update_bin(self):
        bin_timestamp = make_datetime("2011/01/01:05:00:00")
        alert_type = "alert-type"
        alert_name = "alert-name"

        self.ops.update_bin(bin_timestamp, alert_type, alert_name)

        rc_params = self.vals[self.ops.qstr_update_bin]

        self.assertEquals(rc_params, dict(
                timestamp=to_epoch(bin_timestamp),
                type=alert_type,
                name=alert_name))

    def test_commit(self):
        self.assertFalse(self.db_conn.test_committed())
        self.ops.commit()
        self.assertTrue(self.db_conn.test_committed())
